@@ -35,7 +35,6 @@
 (require 'magit-wip)
 
 ;; For `magit-apply'
-(declare-function magit-anti-stage 'magit-rockstar)
 (declare-function magit-am-popup 'magit-sequence)
 ;; For `magit-discard-files'
 (declare-function magit-checkout-stage 'magit)
@@ -56,7 +55,7 @@
   "Whether unstaging a committed change reverts it instead.
 
 A committed change cannot be unstaged, because staging and
-unstaging are actions that are concern with the differences
+unstaging are actions that are concerned with the differences
 between the index and the working tree, not with committed
 changes.
 
@@ -68,12 +67,25 @@ information see command `magit-reverse-in-index'."
   :group 'magit-commands
   :type 'boolean)
 
+(defcustom magit-reverse-atomically nil
+  "Whether to reverse changes atomically.
+
+If some changes can be reversed while others cannot, then nothing
+is reversed if the value of this option is non-nil.  But when it
+is nil, then the changes that can be reversed are reversed and
+for the other changes diff files are created that contain the
+rejected reversals."
+  :package-version '(magit . "2.7.0")
+  :group 'magit-commands
+  :type 'boolean)
+
 ;;; Commands
 ;;;; Apply
 
 (defun magit-apply (&rest args)
-  "Apply the change at point.
-With a prefix argument and if necessary, attempt a 3-way merge."
+  "Apply the change at point to the working tree.
+With a prefix argument fallback to a 3-way merge.  Doing
+so causes the change to be applied to the index as well."
   (interactive (and current-prefix-arg (list "--3way")))
   (--when-let (magit-apply--get-selection)
     (pcase (list (magit-diff-type) (magit-diff-scope))
@@ -87,22 +99,26 @@ With a prefix argument and if necessary, attempt a 3-way merge."
       (`(,_   file) (magit-apply-diff   it args))
       (`(,_  files) (magit-apply-diffs  it args)))))
 
+(defun magit-apply--section-content (section)
+  (buffer-substring-no-properties (if (eq (magit-section-type section) 'hunk)
+                                      (magit-section-start section)
+                                    (magit-section-content section))
+                                  (magit-section-end section)))
+
 (defun magit-apply-diffs (sections &rest args)
   (setq sections (magit-apply--get-diffs sections))
   (magit-apply-patch sections args
                      (mapconcat
                       (lambda (s)
                         (concat (magit-diff-file-header s)
-                                (buffer-substring (magit-section-content s)
-                                                  (magit-section-end s))))
+                                (magit-apply--section-content s)))
                       sections "")))
 
 (defun magit-apply-diff (section &rest args)
   (setq section (car (magit-apply--get-diffs (list section))))
   (magit-apply-patch section args
                      (concat (magit-diff-file-header section)
-                             (buffer-substring (magit-section-content section)
-                                               (magit-section-end section)))))
+                             (magit-apply--section-content section))))
 
 (defun magit-apply-hunks (sections &rest args)
   (let ((section (magit-section-parent (car sections))))
@@ -110,19 +126,15 @@ With a prefix argument and if necessary, attempt a 3-way merge."
       (user-error "Cannot un-/stage resolution hunks.  Stage the whole file"))
     (magit-apply-patch section args
                        (concat (magit-section-diff-header section)
-                               (mapconcat
-                                (lambda (s)
-                                  (buffer-substring (magit-section-start s)
-                                                    (magit-section-end s)))
-                                sections "")))))
+                               (mapconcat 'magit-apply--section-content
+                                          sections "")))))
 
 (defun magit-apply-hunk (section &rest args)
   (when (string-match "^diff --cc" (magit-section-parent-value section))
     (user-error "Cannot un-/stage resolution hunks.  Stage the whole file"))
   (magit-apply-patch (magit-section-parent section) args
                      (concat (magit-diff-file-header section)
-                             (buffer-substring (magit-section-start section)
-                                               (magit-section-end section)))))
+                             (magit-apply--section-content section))))
 
 (defun magit-apply-region (section &rest args)
   (unless (magit-diff-context-p)
@@ -176,12 +188,14 @@ With a prefix argument and if necessary, attempt a 3-way merge."
 
 ;;;; Stage
 
-(defun magit-stage ()
-  "Add the change at point to the staging area."
-  (interactive)
+(defun magit-stage (&optional intent)
+  "Add the change at point to the staging area.
+With a prefix argument, INTENT, and an untracked file (or files)
+at point, stage the file but not its content."
+  (interactive "P")
   (--when-let (magit-apply--get-selection)
     (pcase (list (magit-diff-type) (magit-diff-scope))
-      (`(untracked     ,_) (magit-stage-untracked))
+      (`(untracked     ,_) (magit-stage-untracked intent))
       (`(unstaged  region) (magit-apply-region it "--cached"))
       (`(unstaged    hunk) (magit-apply-hunk   it "--cached"))
       (`(unstaged   hunks) (magit-apply-hunks  it "--cached"))
@@ -233,7 +247,7 @@ ignored) files.
     (mapc #'magit-turn-on-auto-revert-mode-if-desired files))
   (magit-wip-commit-after-apply files " after stage"))
 
-(defun magit-stage-untracked ()
+(defun magit-stage-untracked (&optional intent)
   (let* ((section (magit-current-section))
          (files (pcase (magit-diff-scope)
                   (`file  (list (magit-section-value section)))
@@ -246,7 +260,8 @@ ignored) files.
         (push file plain)))
     (magit-wip-commit-before-change files " before stage")
     (when plain
-      (magit-run-git "add" "--" plain)
+      (magit-run-git "add" (and intent "--intent-to-add")
+                     "--" plain)
       (when magit-auto-revert-mode
         (mapc #'magit-turn-on-auto-revert-mode-if-desired plain)))
     (dolist (repo repos)
@@ -346,7 +361,7 @@ without requiring confirmation."
                (magit-section-parent-value section)))
         (progn (let ((inhibit-magit-refresh t))
                  (funcall apply section "--reverse" "--cached")
-                 (funcall apply section "--reverse"))
+                 (funcall apply section "--reverse" "--reject"))
                (magit-refresh))
       (funcall apply section "--reverse" "--index"))))
 
@@ -367,7 +382,7 @@ without requiring confirmation."
                  (magit-section-parent-value section)))
           (progn (let ((inhibit-magit-refresh t))
                    (funcall apply sections "--reverse" "--cached")
-                   (funcall apply sections "--reverse"))
+                   (funcall apply sections "--reverse" "--reject"))
                  (magit-refresh))
         (funcall apply sections "--reverse" "--index")))))
 
@@ -473,9 +488,10 @@ without requiring confirmation."
             (setq sections
                   (--filter (not (member (magit-section-value it) binaries))
                             sections)))
-          (if (= (length sections) 1)
-              (magit-discard-apply (car sections) 'magit-apply-diff)
-            (magit-discard-apply-n sections 'magit-apply-diffs))
+          (cond ((= (length sections) 1)
+                 (magit-discard-apply (car sections) 'magit-apply-diff))
+                (sections
+                 (magit-discard-apply-n sections 'magit-apply-diffs)))
           (when binaries
             (let ((modified (magit-modified-files t)))
               (setq binaries (--separate (member it modified) binaries)))
@@ -490,7 +506,9 @@ without requiring confirmation."
 ;;;; Reverse
 
 (defun magit-reverse (&rest args)
-  "Reverse the change at point in the working tree."
+  "Reverse the change at point in the working tree.
+With a prefix argument fallback to a 3-way merge.  Doing
+so causes the change to be applied to the index as well."
   (interactive (and current-prefix-arg (list "--3way")))
   (--when-let (magit-apply--get-selection)
     (pcase (list (magit-diff-type) (magit-diff-scope))
@@ -505,18 +523,18 @@ without requiring confirmation."
 
 (defun magit-reverse-region (section args)
   (when (magit-confirm 'reverse "Reverse region")
-    (apply 'magit-apply-region section "--reverse" args)))
+    (magit-reverse-apply section 'magit-apply-region args)))
 
 (defun magit-reverse-hunk (section args)
   (when (magit-confirm 'reverse "Reverse hunk")
-    (apply 'magit-apply-hunk section "--reverse" args)))
+    (magit-reverse-apply section 'magit-apply-hunk args)))
 
 (defun magit-reverse-hunks (sections args)
   (when (magit-confirm 'reverse
           (format "Reverse %s hunks from %s"
                   (length sections)
                   (magit-section-parent-value (car sections))))
-    (magit-apply-hunks sections "--reverse" args)))
+    (magit-reverse-apply sections 'magit-apply-hunks args)))
 
 (defun magit-reverse-file (section args)
   (magit-reverse-files (list section) args))
@@ -527,11 +545,17 @@ without requiring confirmation."
            (--separate (member (magit-section-value it) bs) sections))]
     (when (magit-confirm-files 'reverse (mapcar #'magit-section-value sections))
       (if (= (length sections) 1)
-          (magit-apply-diff (car sections) "--reverse" args)
-        (magit-apply-diffs sections "--reverse" args)))
+          (magit-reverse-apply (car sections) 'magit-apply-diff args)
+        (magit-reverse-apply sections 'magit-apply-diffs args)))
     (when binaries
       (user-error "Cannot reverse binary files"))))
 
+(defun magit-reverse-apply (section:s apply args)
+  (funcall apply section:s "--reverse" args
+           (and (not magit-reverse-atomically)
+                (not (member "--3way" args))
+                "--reject")))
+
 (defun magit-reverse-in-index (&rest args)
   "Reverse the change at point in the index but not the working tree.
 
