@@ -153,7 +153,8 @@ and then turned on again when turning off the latter."
 (defvar-local magit-blame-separator nil)
 
 (define-minor-mode magit-blame-mode
-  "Display blame information inline."
+  "Display blame information inline.
+\n\\{magit-blame-mode-map}"
   :lighter magit-blame-mode-lighter
   (cond (magit-blame-mode
          (when (called-interactively-p 'any)
