@@ -41,6 +41,7 @@
 (declare-function magit-show-commit 'magit)
 (defvar magit-refs-indent-cherry-lines)
 (defvar magit-refs-show-commit-count)
+(defvar magit-status-sections-hook)
 
 (require 'ansi-color)
 (require 'crm)
@@ -171,6 +172,15 @@ This is useful if you use really long branch names."
   "Face for the date part of the log output."
   :group 'magit-faces)
 
+;;;; File Log
+
+(defcustom magit-log-buffer-file-locked t
+  "Whether `magit-log-buffer-file' uses a decicated buffer."
+  :package-version '(magit . "2.7.0")
+  :group 'magit-commands
+  :group 'magit-log
+  :type 'boolean)
+
 ;;;; Select Mode
 
 (defcustom magit-log-select-arguments '("-n256" "--decorate")
@@ -404,6 +414,25 @@ the upstream isn't ahead of the current branch) show."
            (default-value 'magit-log-arguments))))
     (magit-invoke-popup 'magit-log-popup nil arg)))
 
+;;;###autoload
+(defun magit-log-buffer-file-popup ()
+  "Popup console for log commands.
+
+This is a variant of `magit-log-popup' which shows the same popup
+but which limits the log to the file being visited in the current
+buffer."
+  (interactive)
+  (-if-let (file (magit-file-relative-name))
+      (let ((magit-log-arguments
+             (magit-popup-import-file-args
+              (-if-let (buffer (magit-mode-get-buffer 'magit-log-mode))
+                  (with-current-buffer buffer
+                    (nth 2 magit-refresh-args))
+                (default-value 'magit-log-arguments))
+              (list file))))
+        (magit-invoke-popup 'magit-log-popup nil nil))
+    (user-error "Buffer isn't visiting a file")))
+
 (defun magit-log-refresh-popup (arg)
   "Popup console for changing log arguments in the current buffer."
   (interactive "P")
@@ -572,19 +601,22 @@ With a prefix argument or when `--follow' is part of
                          (1- (line-number-at-pos (region-end))))
                  (list current-prefix-arg)))
   (-if-let (file (magit-file-relative-name))
-      (magit-mode-setup #'magit-log-mode
-                        (list (or magit-buffer-refname
-                                  (magit-get-current-branch) "HEAD"))
-                        (let ((args (car (magit-log-arguments))))
-                          (when (and follow (not (member "--follow" args)))
-                            (push "--follow" args))
-                          (when (and beg end)
-                            (setq args (cons (format "-L%s,%s:%s" beg end file)
-                                             (cl-delete "-L" args :test
-                                                        'string-prefix-p)))
-                            (setq file nil))
-                          args)
-                        (and file (list file)))
+      (magit-mode-setup-internal
+       #'magit-log-mode
+       (list (list (or magit-buffer-refname
+                       (magit-get-current-branch)
+                       "HEAD"))
+             (let ((args (car (magit-log-arguments))))
+               (when (and follow (not (member "--follow" args)))
+                 (push "--follow" args))
+               (when (and beg end)
+                 (setq args (cons (format "-L%s,%s:%s" beg end file)
+                                  (cl-delete "-L" args :test
+                                             'string-prefix-p)))
+                 (setq file nil))
+               args)
+             (and file (list file)))
+       magit-log-buffer-file-locked)
     (user-error "Buffer isn't visiting a file"))
   (magit-log-goto-same-commit))
 
@@ -878,8 +910,8 @@ Do not add this to a hook variable."
           (insert ?\s))
         (when side
           (insert (propertize side 'face (if (string= side "<")
-                                             'magit-diff-removed
-                                           'magit-diff-added)))
+                                             'magit-cherry-equivalent
+                                           'magit-cherry-unmatched)))
           (insert ?\s))
         (when align
           (insert (propertize hash 'face 'magit-hash) ?\s))
@@ -1294,8 +1326,12 @@ Type \\[magit-reset] to reset HEAD to the commit at point.
 (defun magit-insert-unpulled-from-pushremote ()
   "Insert commits that haven't been pulled from the push-remote yet."
   (--when-let (magit-get-push-branch)
-    (unless (equal (magit-rev-name it)
-                   (magit-rev-name "@{upstream}"))
+    (unless (and (equal (magit-rev-name it)
+                        (magit-rev-name "@{upstream}"))
+                 (or (memq 'magit-insert-unpulled-from-upstream
+                           magit-status-sections-hook)
+                     (memq 'magit-insert-unpulled-from-upstream-or-recent
+                           magit-status-sections-hook)))
       (magit-insert-section (unpulled (concat ".." it))
         (magit-insert-heading
           (format (propertize "Unpulled from %s:" 'face 'magit-section-heading)
@@ -1329,8 +1365,10 @@ Type \\[magit-reset] to reset HEAD to the commit at point.
 (defun magit-insert-unpushed-to-pushremote ()
   "Insert commits that haven't been pushed to the push-remote yet."
   (--when-let (magit-get-push-branch)
-    (unless (equal (magit-rev-name it)
-                   (magit-rev-name "@{upstream}"))
+    (unless (and (equal (magit-rev-name it)
+                        (magit-rev-name "@{upstream}"))
+                 (memq 'magit-insert-unpushed-to-upstream
+                       magit-status-sections-hook))
       (magit-insert-section (unpushed (concat it ".."))
         (magit-insert-heading
           (format (propertize "Unpushed to %s:" 'face 'magit-section-heading)
@@ -1357,10 +1395,11 @@ ahead of the current branch, then show the commits that have
 not yet been pulled into the current branch.  If no upstream is
 configured or if the upstream is not ahead of the current branch,
 then show the last `magit-log-section-commit-count' commits."
-  (if (equal (magit-rev-parse "HEAD")
-             (magit-rev-parse "@{upstream}"))
-      (magit-insert-recent-commits t)
-    (magit-insert-unpulled-from-upstream)))
+  (let ((upstream (magit-rev-parse "@{upstream}")))
+    (if (or (not upstream)
+            (equal upstream (magit-rev-parse "HEAD")))
+        (magit-insert-recent-commits t)
+      (magit-insert-unpulled-from-upstream))))
 
 ;;;; Auxiliary Log Sections
 
