@@ -330,6 +330,7 @@ starts complicating other things, then it will be removed."
     (define-key map "m" 'magit-merge-popup)
     (define-key map "M" 'magit-remote-popup)
     (define-key map "o" 'magit-submodule-popup)
+    (define-key map "O" 'magit-subtree-popup)
     (define-key map "P" 'magit-push-popup)
     (define-key map "r" 'magit-rebase-popup)
     (define-key map "R" 'magit-file-rename)
@@ -349,6 +350,7 @@ starts complicating other things, then it will be removed."
     (define-key map "w" 'magit-am-popup)
     (define-key map "W" 'magit-patch-popup)
     (define-key map "x" 'magit-reset)
+    (define-key map "X" 'magit-reset-popup)
     (define-key map "y" 'magit-show-refs-popup)
     (define-key map "Y" 'magit-cherry)
     (define-key map "z" 'magit-stash-popup)
@@ -480,7 +482,15 @@ Magit is documented in info node `(magit)'."
 
 (defun magit-mode-setup (mode &rest args)
   "Setup up a MODE buffer using ARGS to generate its content."
-  (let ((buffer (magit-mode-get-buffer mode t))
+  (magit-mode-setup-internal mode args))
+
+(defun magit-mode-setup-internal (mode args &optional locked)
+  "Setup up a MODE buffer using ARGS to generate its content.
+When optional LOCKED is non-nil, then create a buffer that is
+locked to its value, which is derived from MODE and ARGS."
+  (let ((buffer (magit-mode-get-buffer
+                 mode t nil
+                 (and locked (magit-buffer-lock-value mode args))))
         (section (magit-current-section)))
     (with-current-buffer buffer
       (setq magit-previous-section section)
@@ -553,26 +563,30 @@ thinking a buffer belongs to a repo that it doesn't.")
 (defvar-local magit-buffer-locked-p nil)
 (put 'magit-buffer-locked-p 'permanent-local t)
 
-(defun magit-mode-get-buffer (mode &optional create frame)
+(defun magit-mode-get-buffer (mode &optional create frame value)
   (-if-let (topdir (magit-toplevel))
       (or (--first (with-current-buffer it
                      (and (eq major-mode mode)
                           (equal magit--default-directory topdir)
-                          (not magit-buffer-locked-p)))
+                          (if value
+                              (and magit-buffer-locked-p
+                                   (equal (magit-buffer-lock-value) value))
+                            (not magit-buffer-locked-p))))
                    (if frame
                        (-map #'window-buffer
                              (window-list (unless (eq frame t) frame)))
                      (buffer-list)))
           (and create
                (let ((default-directory topdir))
-                 (magit-generate-new-buffer mode))))
+                 (magit-generate-new-buffer mode value))))
     (user-error "Not inside a Git repository")))
 
-(defun magit-generate-new-buffer (mode)
-  (let* ((name (funcall magit-generate-buffer-name-function mode))
+(defun magit-generate-new-buffer (mode &optional value)
+  (let* ((name (funcall magit-generate-buffer-name-function mode value))
          (buffer (generate-new-buffer name)))
     (with-current-buffer buffer
-      (setq magit--default-directory default-directory))
+      (setq magit--default-directory default-directory)
+      (setq magit-buffer-locked-p (and value t)))
     (when magit-uniquify-buffer-names
       (add-to-list 'uniquify-list-buffers-directory-modes mode)
       (with-current-buffer buffer
@@ -620,33 +634,53 @@ latter is displayed in its place."
   (if magit-buffer-locked-p
       (-if-let (unlocked (magit-mode-get-buffer major-mode))
           (let ((locked (current-buffer)))
-            (set-buffer unlocked)
+            (switch-to-buffer unlocked nil t)
             (kill-buffer locked))
         (setq magit-buffer-locked-p nil)
         (rename-buffer (funcall magit-generate-buffer-name-function
                                 major-mode)))
-    (setq magit-buffer-locked-p
-          (cond ((memq major-mode '(magit-cherry-mode
-                                    magit-log-mode
-                                    magit-reflog-mode
-                                    magit-refs-mode
-                                    magit-revision-mode
-                                    magit-stash-mode
-                                    magit-stashes-mode))
-                 (car magit-refresh-args))
-                ((eq major-mode 'magit-diff-mode)
-                 (let ((rev  (nth 0 magit-refresh-args))
-                       (args (nth 1 magit-refresh-args)))
-                   (cond
-                    ((member "--no-index" args)
-                     (nth 3 magit-refresh-args))
-                    (rev (if args (cons rev args) rev))
-                    (t   (if (member "--cached" args) "staged" "unstaged")))))))
-    (if magit-buffer-locked-p
-        (rename-buffer (funcall magit-generate-buffer-name-function
-                                major-mode magit-buffer-locked-p))
+    (-if-let (value (magit-buffer-lock-value))
+        (-if-let (locked (magit-mode-get-buffer major-mode nil nil value))
+            (let ((unlocked (current-buffer)))
+              (switch-to-buffer locked nil t)
+              (kill-buffer unlocked))
+          (setq magit-buffer-locked-p t)
+          (rename-buffer (funcall magit-generate-buffer-name-function
+                                  major-mode value)))
       (user-error "Buffer has no value it could be locked to"))))
 
+(cl-defun magit-buffer-lock-value
+    (&optional (mode major-mode)
+               (args magit-refresh-args))
+  (cl-case mode
+    (magit-cherry-mode
+     (-let [(upstream head) args]
+       (concat head ".." upstream)))
+    (magit-diff-mode
+     (-let [(rev-or-range const _args files) args]
+       (nconc (cons (or rev-or-range
+                        (if (member "--cached" const)
+                            (progn (setq const (delete "--cached" const))
+                                   'staged)
+                          'unstaged))
+                    const)
+              (and files (cons "--" files)))))
+    (magit-log-mode
+     (-let [(revs _args files) args]
+       (if (and revs files)
+           (append revs (cons "--" files))
+         (append revs files))))
+    (magit-refs-mode
+     (-let [(ref args) args]
+       (cons (or ref "HEAD") args)))
+    (magit-revision-mode
+     (-let [(rev __const _args files) args]
+       (if files (cons rev files) (list rev))))
+    ((magit-reflog-mode   ; (ref ~args)
+      magit-stash-mode    ; (stash _const _args _files)
+      magit-stashes-mode) ; (ref)
+     (car args))))
+
 (defun magit-mode-bury-buffer (&optional kill-buffer)
   "Bury the current buffer.
 With a prefix argument, kill the buffer instead.
@@ -690,16 +724,26 @@ Refresh the current buffer if its major mode derives from
 Run hooks `magit-pre-refresh-hook' and `magit-post-refresh-hook'."
   (interactive)
   (unless inhibit-magit-refresh
-    (magit-run-hook-with-benchmark 'magit-pre-refresh-hook)
-    (when (derived-mode-p 'magit-mode)
-      (magit-refresh-buffer))
-    (--when-let (and magit-refresh-status-buffer
-                     (not (derived-mode-p 'magit-status-mode))
-                     (magit-mode-get-buffer 'magit-status-mode))
-      (with-current-buffer it
-        (magit-refresh-buffer)))
-    (magit-auto-revert-buffers)
-    (magit-run-hook-with-benchmark 'magit-post-refresh-hook)))
+    (let ((start (current-time))
+          (magit--refresh-cache (list (cons 0 0))))
+      (when magit-refresh-verbose
+        (message "Refreshing magit..."))
+      (magit-run-hook-with-benchmark 'magit-pre-refresh-hook)
+      (when (derived-mode-p 'magit-mode)
+        (magit-refresh-buffer))
+      (--when-let (and magit-refresh-status-buffer
+                       (not (derived-mode-p 'magit-status-mode))
+                       (magit-mode-get-buffer 'magit-status-mode))
+        (with-current-buffer it
+          (magit-refresh-buffer)))
+      (magit-auto-revert-buffers)
+      (magit-run-hook-with-benchmark 'magit-post-refresh-hook)
+      (when magit-refresh-verbose
+        (message "Refreshing magit...done (%.3fs, cached %s/%s)"
+                 (float-time (time-subtract (current-time) start))
+                 (caar magit--refresh-cache)
+                 (+ (caar magit--refresh-cache)
+                    (cdar magit--refresh-cache)))))))
 
 (defun magit-refresh-all ()
   "Refresh all buffers belonging to the current repository.
@@ -722,7 +766,8 @@ Run hooks `magit-pre-refresh-hook' and `magit-post-refresh-hook'."
   "Refresh the current Magit buffer."
   (setq magit-refresh-start-time (current-time))
   (let ((refresh (intern (format "%s-refresh-buffer"
-                                 (substring (symbol-name major-mode) 0 -5)))))
+                                 (substring (symbol-name major-mode) 0 -5))))
+        (magit--refresh-cache (or magit--refresh-cache (list (cons 0 0)))))
     (when (functionp refresh)
       (when magit-refresh-verbose
         (message "Refreshing buffer `%s'..." (buffer-name)))
@@ -801,13 +846,15 @@ Run hooks `magit-pre-refresh-hook' and `magit-post-refresh-hook'."
 This function is intended to be added to `after-save-hook'.
 
 If the status buffer does not exist or the file being visited in
-the current buffer isn't inside a repository, then do nothing.
+the current buffer isn't inside the working tree of a repository,
+then do nothing.
 
 Note that refreshing a Magit buffer is done by re-creating its
 contents from scratch, which can be slow in large repositories.
 If you are not satisfied with Magit's performance, then you
 should obviously not add this function to that hook."
-  (unless disable-magit-save-buffers
+  (when (and (not disable-magit-save-buffers)
+             (magit-inside-worktree-p))
     (--when-let (ignore-errors (magit-mode-get-buffer 'magit-status-mode))
       (add-to-list 'magit-after-save-refresh-buffers it)
       (add-hook 'post-command-hook 'magit-after-save-refresh-buffers))))
