@@ -172,6 +172,12 @@ non-nil, then the password is read from the user instead."
   :group 'magit-process
   :type '(repeat (regexp)))
 
+(defcustom magit-process-ensure-unix-line-ending t
+  "Whether Magit should ensure a unix coding system when talking to Git."
+  :package-version '(magit . "2.6.0")
+  :group 'magit-process
+  :type 'boolean)
+
 (defface magit-process-ok
   '((t :inherit magit-section-heading :foreground "green"))
   "Face for zero exit-status."
@@ -301,9 +307,11 @@ Process output goes into a new section in the buffer returned by
 (defun magit-process-file (&rest args)
   "Process files synchronously in a separate process.
 Identical to `process-file' but temporarily enable Cygwin's
-\"noglob\" option during the call."
+\"noglob\" option during the call and ensure unix eol
+conversion."
   (let ((process-environment (append (magit-cygwin-env-vars)
-                                     process-environment)))
+                                     process-environment))
+        (default-process-coding-system (magit--process-coding-system)))
     (apply #'process-file args)))
 
 (defun magit-cygwin-env-vars ()
@@ -337,11 +345,18 @@ flattened before use."
         (while (and magit-this-process
                     (eq (process-status magit-this-process) 'run))
           (sleep-for 0.005)))
-    (let ((process-environment (append (magit-cygwin-env-vars)
-                                       process-environment)))
-      (apply #'call-process-region (point-min) (point-max)
-             magit-git-executable nil nil nil
-             (magit-process-git-arguments args)))))
+    (run-hooks 'magit-pre-call-git-hook)
+    (-let* ((process-environment (append (magit-cygwin-env-vars)
+                                         process-environment))
+            (default-process-coding-system (magit--process-coding-system))
+            (flat-args (magit-process-git-arguments args))
+            ((process-buf . section)
+             (magit-process-setup magit-git-executable flat-args))
+            (inhibit-read-only t))
+      (magit-process-finish
+       (apply #'call-process-region (point-min) (point-max)
+              magit-git-executable nil process-buf nil flat-args)
+       process-buf nil default-directory section))))
 
 (defun magit-run-git-with-logfile (file &rest args)
   "Call Git in a separate process and log its output to FILE.
@@ -454,7 +469,8 @@ Magit status buffer."
                   ;; which would modify the input (issue #20).
                   (and (not input) magit-process-connection-type))
                  (process-environment (append (magit-cygwin-env-vars)
-                                              process-environment)))
+                                              process-environment))
+                 (default-process-coding-system (magit--process-coding-system)))
              (apply #'start-file-process
                     (file-name-nondirectory program)
                     process-buf program args))))
@@ -666,6 +682,14 @@ Return the matched string suffixed with \": \", if needed."
             ((string-suffix-p ":"  prompt) (concat prompt " "))
             (t                             (concat prompt ": "))))))
 
+(defun magit--process-coding-system ()
+  (if magit-process-ensure-unix-line-ending
+      (cons (coding-system-change-eol-conversion
+             (car default-process-coding-system) 'unix)
+            (coding-system-change-eol-conversion
+             (cdr default-process-coding-system) 'unix))
+      default-process-coding-system))
+
 (defvar magit-credential-hook nil
   "Hook run before Git needs credentials.")
 
@@ -787,10 +811,10 @@ as argument."
                             (--when-let (magit-section-content section)
                               (when (re-search-backward
                                      magit-process-error-message-re it t)
-                                (match-string 1))))))
+                                (match-string-no-properties 1))))))
                    "Git failed")))
       (if magit-process-raise-error
-          (signal 'magit-git-error msg)
+          (signal 'magit-git-error (list (format "%s (in %s)" msg default-dir)))
         (--when-let (magit-mode-get-buffer 'magit-status-mode)
           (setq magit-this-error msg))
         (message "%s ... [%s buffer %s for details]" msg
