@@ -61,20 +61,28 @@ Then show the status buffer for the new repository."
                 (and (string-match "\\([^./]+\\)\\(\\.git\\)?$" url)
                      (match-string 1 url))))))
   (setq directory (file-name-as-directory (expand-file-name directory)))
-  (message "Cloning %s..." repository)
-  (when (= (magit-call-git "clone" repository
-                           ;; Stop cygwin git making a "c:" directory.
-                           (magit-convert-git-filename directory))
-           0)
-    (when (or (eq  magit-clone-set-remote.pushDefault t)
-              (and magit-clone-set-remote.pushDefault
-                   (y-or-n-p "Set `remote.pushDefault' to \"origin\"? ")))
-      (let ((default-directory directory))
-        (magit-call-git "config" "remote.pushDefault" "origin")))
-    (unless magit-clone-set-remote-head
-      (magit-remote-unset-head "origin"))
-    (message "Cloning %s...done" repository)
-    (magit-status-internal directory)))
+  (magit-run-git-async "clone" repository
+                       ;; Stop cygwin git making a "c:" directory.
+                       (magit-convert-git-filename directory))
+  ;; Don't refresh the buffer we're calling from.
+  (process-put magit-this-process 'inhibit-refresh t)
+  (set-process-sentinel
+   magit-this-process
+   (lambda (process event)
+     (when (memq (process-status process) '(exit signal))
+       (let ((magit-process-raise-error t))
+         (magit-process-sentinel process event)))
+     (when (and (eq (process-status process) 'exit)
+                (= (process-exit-status process) 0))
+       (let ((default-directory directory))
+         (when (or (eq  magit-clone-set-remote.pushDefault t)
+                   (and magit-clone-set-remote.pushDefault
+                        (y-or-n-p "Set `remote.pushDefault' to \"origin\"? ")))
+           (magit-call-git "config" "remote.pushDefault" "origin"))
+         (unless magit-clone-set-remote-head
+           (magit-remote-unset-head "origin")))
+       (with-current-buffer (process-get process 'command-buf)
+         (magit-status-internal directory))))))
 
 ;;; Setup
 
@@ -178,12 +186,16 @@ Delete the symbolic-ref \"refs/remotes/<remote>/HEAD\"."
   'magit-commands
   :man-page "git-fetch"
   :switches '((?p "Prune deleted branches" "--prune"))
-  :actions  '("Fetch from"
+  :actions  '("Configure"
+              (?C "variables..."           magit-branch-config-popup)
+              "Fetch from"
               (?p magit-get-push-remote    magit-fetch-from-pushremote)
               (?u magit-get-remote         magit-fetch-from-upstream)
               (?e "elsewhere"              magit-fetch)
               (?a "all remotes"            magit-fetch-all)
               "Fetch"
+              (?o "another branch"         magit-fetch-branch)
+              (?r "explicit refspec"       magit-fetch-refspec)
               (?m "submodules"             magit-submodule-fetch))
   :default-action 'magit-fetch
   :max-action-columns 1)
@@ -219,6 +231,26 @@ Delete the symbolic-ref \"refs/remotes/<remote>/HEAD\"."
                      (magit-fetch-arguments)))
   (magit-git-fetch remote args))
 
+;;;###autoload
+(defun magit-fetch-branch (remote branch args)
+  "Fetch a BRANCH from a REMOTE."
+  (interactive
+   (let ((remote (magit-read-remote-or-url "Fetch from remote or url")))
+     (list remote
+           (magit-read-remote-branch "Fetch branch" remote)
+           (magit-fetch-arguments))))
+  (magit-git-fetch remote (cons branch args)))
+
+;;;###autoload
+(defun magit-fetch-refspec (remote refspec args)
+  "Fetch a REFSPEC from a REMOTE."
+  (interactive
+   (let ((remote (magit-read-remote-or-url "Fetch from remote or url")))
+     (list remote
+           (magit-read-refspec "Fetch using refspec" remote)
+           (magit-fetch-arguments))))
+  (magit-git-fetch remote (cons refspec args)))
+
 ;;;###autoload
 (defun magit-fetch-all (args)
   "Fetch from all remotes."
@@ -249,10 +281,11 @@ removed on the respective remote."
   "Popup console for pull commands."
   'magit-commands
   :man-page "git-pull"
-  :variables '("Variables"
+  :variables '("Configure"
                (?r "branch.%s.rebase"
                    magit-cycle-branch*rebase
-                   magit-pull-format-branch*rebase))
+                   magit-pull-format-branch*rebase)
+               (?C "variables..." magit-branch-config-popup))
   :actions '((lambda ()
                (--if-let (magit-get-current-branch)
                    (concat
@@ -294,10 +327,11 @@ missing.  To add them use something like:
       \\='magit-fetch-from-push-remote ?F))"
   'magit-commands
   :man-page "git-pull"
-  :variables '("Pull variables"
+  :variables '("Configure"
                (?r "branch.%s.rebase"
                    magit-cycle-branch*rebase
-                   magit-pull-format-branch*rebase))
+                   magit-pull-format-branch*rebase)
+               (?C "variables..." magit-branch-config-popup))
   :actions '((lambda ()
                (--if-let (magit-get-current-branch)
                    (concat
@@ -312,6 +346,8 @@ missing.  To add them use something like:
              (?f "remotes"           magit-fetch-all-no-prune)
              (?F "remotes and prune" magit-fetch-all-prune)
              "Fetch"
+             (?o "another branch"    magit-fetch-branch)
+             (?s "explicit refspec"  magit-fetch-refspec)
              (?m "submodules"        magit-submodule-fetch))
   :default-action 'magit-fetch
   :max-action-columns 1)
@@ -343,7 +379,9 @@ missing.  To add them use something like:
   "Pull from the upstream of the current branch."
   (interactive (list (magit-pull-arguments)))
   (--if-let (magit-get-upstream-branch)
-      (magit-git-pull it args)
+      (progn (run-hooks 'magit-credential-hook)
+             (magit-run-git-with-editor
+              "pull" args (car (magit-split-branch-name it))))
     (--if-let (magit-get-current-branch)
         (user-error "No upstream is configured for %s" it)
       (user-error "No branch is checked out"))))
@@ -377,9 +415,11 @@ available in the popup.  If the value is t, then that argument is
 redundant.  But note that changing the value of this option does
 not take affect immediately, the argument will only be added or
 removed after restarting Emacs."
-  :package-version '(magit . "2.4.0")
+  :package-version '(magit . "2.6.0")
   :group 'magit-commands
-  :type 'boolean)
+  :type '(choice (const :tag "don't set" nil)
+                 (const :tag "set branch.<name>.pushRemote" t)
+                 (const :tag "set remote.pushDefault" default)))
 
 ;;;###autoload (autoload 'magit-push-popup "magit-remote" nil t)
 (magit-define-popup magit-push-popup
@@ -391,7 +431,9 @@ removed after restarting Emacs."
               (?d "Dry run"       "--dry-run")
               ,@(and (not magit-push-current-set-remote-if-missing)
                      '((?u "Set upstream"  "--set-upstream"))))
-  :actions '((lambda ()
+  :actions '("Configure"
+             (?C "variables..."      magit-branch-config-popup)
+             (lambda ()
                (--when-let (magit-get-current-branch)
                  (concat (propertize "Push " 'face 'magit-popup-heading)
                          (propertize it      'face 'magit-branch-local)
@@ -404,8 +446,9 @@ removed after restarting Emacs."
              "Push"
              (?o "another branch"    magit-push)
              (?T "a tag"             magit-push-tag)
-             (?m "matching branches" magit-push-matching)
-             (?t "all tags"          magit-push-tags))
+             (?r "explicit refspecs" magit-push-refspecs)
+             (?t "all tags"          magit-push-tags)
+             (?m "matching branches" magit-push-matching))
   :max-action-columns 2)
 
 (defun magit-git-push (branch target args)
@@ -427,14 +470,20 @@ the push-remote can be changed before pushed to it."
   (interactive
    (list (magit-push-arguments)
          (and (magit--push-current-set-pushremote-p current-prefix-arg)
-              (magit-read-remote (format "Set push-remote of %s and push there"
-                                         (magit-get-current-branch))))))
+              (magit-read-remote
+               (if (eq magit-push-current-set-remote-if-missing 'default)
+                   "Set `remote.pushDefault' and push there"
+                 (format "Set `branch.%s.pushRemote' and push there"
+                         (magit-get-current-branch)))))))
   (--if-let (magit-get-current-branch)
       (progn (when push-remote
-               (magit-call-git "config"
-                               (format "branch.%s.pushRemote"
-                                       (magit-get-current-branch))
-                               push-remote))
+               (magit-call-git
+                "config"
+                (if (eq magit-push-current-set-remote-if-missing 'default)
+                    "remote.pushDefault"
+                  (format "branch.%s.pushRemote"
+                          (magit-get-current-branch)))
+                push-remote))
              (-if-let (remote (magit-get-push-remote it))
                  (if (member remote (magit-list-remotes))
                      (magit-git-push it (concat remote "/" it) args)
@@ -452,8 +501,12 @@ the push-remote can be changed before pushed to it."
   (--if-let (magit-get-push-branch)
       (concat (magit-branch-set-face it) "\n")
     (and (magit--push-current-set-pushremote-p)
-         (concat (propertize "pushRemote" 'face 'bold)
-                 ", after setting that\n"))))
+         (concat
+          (propertize (if (eq magit-push-current-set-remote-if-missing 'default)
+                          "pushDefault"
+                        "pushRemote")
+                      'face 'bold)
+          ", after setting that\n"))))
 
 ;;;###autoload
 (defun magit-push-current-to-upstream (args &optional upstream)
@@ -513,6 +566,22 @@ Both the source and the target are read in the minibuffer."
            (magit-push-arguments))))
   (magit-git-push source target args))
 
+;;;###autoload
+(defun magit-push-refspecs (remote refspecs args)
+  "Push one or multiple REFSPECS to a REMOTE.
+Both the REMOTE and the REFSPECS are read in the minibuffer.  To
+use multiple REFSPECS, separate them with commas.  Completion is
+only available for the part before the colon, or when no colon
+is used."
+  (interactive
+   (list (magit-read-remote "Push to remote")
+         (completing-read-multiple
+          "Push refspec,s: "
+          (cons "HEAD" (magit-list-local-branch-names)))
+         (magit-push-arguments)))
+  (run-hooks 'magit-credential-hook)
+  (magit-run-git-async "push" "-v" args remote refspecs))
+
 ;;;###autoload
 (defun magit-push-matching (remote &optional args)
   "Push all matching branches to another repository.
