@@ -41,29 +41,40 @@
               (?d "Deinit" magit-submodule-deinit)))
 
 ;;;###autoload
-(defun magit-submodule-add (url &optional path)
+(defun magit-submodule-add (url &optional path name)
   "Add the repository at URL as a submodule.
+
 Optional PATH is the path to the submodule relative to the root
-of the superproject. If it is nil then the path is determined
-based on URL."
+of the superproject.  If it is nil, then the path is determined
+based on the URL.
+
+Optional NAME is the name of the submodule.  If it is nil, then
+PATH also becomes the name."
   (interactive
    (magit-with-toplevel
-     (let ((path (read-file-name
-                  "Add submodule: " nil nil nil
-                  (magit-section-when [file untracked]
-                    (directory-file-name (magit-section-value it))))))
-       (when path
-         (setq path (file-name-as-directory (expand-file-name path)))
-         (when (member path (list "" default-directory))
-           (setq path nil)))
+     (let ((path (magit-completing-read "Add submodules at path"
+                                        (magit-untracked-files)
+                                        #'magit-git-repo-p nil nil nil
+                                        (magit-section-when [file untracked]
+                                          (file-relative-name
+                                           (magit-section-value it)
+                                           default-directory)))))
+       (unless path
+         (user-error "No path selected"))
        (list (magit-read-string-ns
               "Remote url"
-              (and path (magit-git-repo-p path t)
-                   (let ((default-directory path))
-                     (magit-get "remote" (or (magit-get-remote) "origin")
-                                "url"))))
-             (and path (directory-file-name (file-relative-name path)))))))
-  (magit-run-git "submodule" "add" url path))
+              (let ((default-directory (file-name-as-directory
+                                        (expand-file-name path))))
+                (magit-get "remote" (or (magit-get-remote) "origin") "url")))
+             (directory-file-name path)
+             (magit-read-string-ns
+              "Name submodule" nil nil
+              (or (--keep (-let [(var val) (split-string it "=")]
+                            (and (equal val path)
+                                 (cadr (split-string var "\\."))))
+                          (magit-git-lines "config" "--list" "-f" ".gitmodules"))
+                  (directory-file-name path)))))))
+  (magit-run-git "submodule" "add" (and name (list "--name" name)) url path))
 
 ;;;###autoload
 (defun magit-submodule-setup ()
@@ -114,59 +125,94 @@ With a prefix argument fetch all remotes."
 ;;; Sections
 
 ;;;###autoload
-(defun magit-insert-submodule-commits (section range)
+(defun magit-insert-submodules ()
+  "Insert sections for all modules.
+For each section insert the path and the output of `git describe --tags'."
+  (-when-let (modules (magit-get-submodules))
+    (magit-insert-section (modules nil t)
+      (magit-insert-heading "Modules:")
+      (magit-with-toplevel
+        (dolist (module modules)
+          (let ((default-directory
+                  (expand-file-name (file-name-as-directory module))))
+            (magit-insert-section (file module t)
+              (insert (format "%-25s " module))
+              (--when-let (magit-git-string "describe" "--tags")
+                (when (string-match-p "\\`[0-9]" it)
+                  (insert ?\s))
+                (insert it))
+              (insert ?\n)))))
+      (insert ?\n))))
+
+;;;###autoload
+(defun magit-insert-modules-unpulled-from-upstream ()
+  "Insert sections for modules that haven't been pulled from the upstream.
+These sections can be expanded to show the respective commits."
+  (magit--insert-modules-logs "Modules unpulled from @{upstream}"
+                              'modules-unpulled-from-upstream
+                              'magit-get-upstream-ref
+                              "HEAD..%s"))
+
+;;;###autoload
+(defun magit-insert-modules-unpulled-from-pushremote ()
+  "Insert sections for modules that haven't been pulled from the push-remote.
+These sections can be expanded to show the respective commits."
+  (magit--insert-modules-logs "Modules unpulled from <push-remote>"
+                              'modules-unpulled-from-pushremote
+                              'magit-get-push-branch
+                              "HEAD..%s"))
+
+;;;###autoload
+(defun magit-insert-modules-unpushed-to-upstream ()
+  "Insert sections for modules that haven't been pushed to the upstream.
+These sections can be expanded to show the respective commits."
+  (magit--insert-modules-logs "Modules unmerged into @{upstream}"
+                              'modules-unpushed-to-upstream
+                              'magit-get-upstream-ref
+                              "%s..HEAD"))
+
+;;;###autoload
+(defun magit-insert-modules-unpushed-to-pushremote ()
+  "Insert sections for modules that haven't been pushed to the push-remote.
+These sections can be expanded to show the respective commits."
+  (magit--insert-modules-logs "Modules unpushed to <push-remote>"
+                              'modules-unpushed-to-pushremote
+                              'magit-get-push-branch
+                              "%s..HEAD"))
+
+(defun magit--insert-modules-logs (heading type fn format)
   "For internal use, don't add to a hook."
-  (if (magit-section-hidden section)
-      (setf (magit-section-washer section)
-            (apply-partially #'magit-insert-submodule-commits section range))
-    (magit-git-wash (apply-partially 'magit-log-wash-log 'module)
-      "log" "--oneline" range)
-    (when (> (point) (magit-section-content section))
-      (delete-char -1))))
-
-;;;###autoload
-(defun magit-insert-unpulled-module-commits ()
-  "Insert sections for all submodules with unpulled commits.
-These sections can be expanded to show the respective commits."
   (-when-let (modules (magit-get-submodules))
-    (magit-insert-section section (unpulled-modules)
-      (magit-insert-heading "Unpulled modules:")
+    (magit-insert-section section ((eval type) nil t)
+      (string-match "\\`\\(.+\\) \\([^ ]+\\)\\'" heading)
+      (magit-insert-heading
+        (concat
+         (propertize (match-string 1 heading) 'face 'magit-section-heading) " "
+         (propertize (match-string 2 heading) 'face 'magit-branch-remote) ":"))
       (magit-with-toplevel
         (dolist (module modules)
           (let ((default-directory
                   (expand-file-name (file-name-as-directory module))))
-            (-when-let (tracked (magit-get-upstream-ref))
+            (--when-let (and (magit-file-accessible-directory-p default-directory)
+                             (funcall fn))
               (magit-insert-section sec (file module t)
                 (magit-insert-heading
                   (concat (propertize module 'face 'magit-diff-file-heading) ":"))
-                (magit-insert-submodule-commits
-                 section (concat "HEAD.." tracked)))))))
-      (if (> (point) (magit-section-content section))
-          (insert ?\n)
-        (magit-cancel-section)))))
-
-;;;###autoload
-(defun magit-insert-unpushed-module-commits ()
-  "Insert sections for all submodules with unpushed commits.
-These sections can be expanded to show the respective commits."
-  (-when-let (modules (magit-get-submodules))
-    (magit-insert-section section (unpushed-modules)
-      (magit-insert-heading "Unpushed modules:")
-      (magit-with-toplevel
-        (dolist (module modules)
-          (let ((default-directory
-                  (expand-file-name (file-name-as-directory module))))
-            (-when-let (tracked (magit-get-upstream-ref))
-              (magit-insert-section sec (file module t)
-                (magit-insert-heading
-                  (concat (propertize module 'face 'magit-diff-file-heading) ":"))
-                (magit-insert-submodule-commits
-                 section (concat tracked "..HEAD")))))))
+                (magit-git-wash (apply-partially 'magit-log-wash-log 'module)
+                  "log" "--oneline" (format format it))
+                (when (> (point) (magit-section-content sec))
+                  (delete-char -1)))))))
       (if (> (point) (magit-section-content section))
           (insert ?\n)
         (magit-cancel-section)))))
 
 ;;; magit-submodule.el ends soon
+
+(define-obsolete-function-alias 'magit-insert-unpulled-module-commits
+  'magit-insert-modules-unpulled-from-upstream "Magit 2.6.0")
+(define-obsolete-function-alias 'magit-insert-unpushed-module-commits
+  'magit-insert-modules-unpushed-to-upstream "Magit 2.6.0")
+
 (provide 'magit-submodule)
 ;; Local Variables:
 ;; indent-tabs-mode: nil
