@@ -207,13 +207,13 @@ commit message."
     (unless (equal parent wipref)
       (setq start-msg (concat "restart autosaving " start-msg))
       (magit-update-ref wipref start-msg
-                        (magit-git-string "commit-tree" "-p" parent
-                                          "-m" start-msg
+                        (magit-git-string "commit-tree" "--no-gpg-sign"
+                                          "-p" parent "-m" start-msg
                                           (concat parent "^{tree}")))
       (setq parent wipref))
     (magit-update-ref wipref msg
-                      (magit-git-string "commit-tree" tree
-                                        "-p" parent "-m" msg))))
+                      (magit-git-string "commit-tree" "--no-gpg-sign"
+                                        "-p" parent "-m" msg tree))))
 
 (defun magit-wip-get-ref ()
   (let ((ref (or (magit-git-string "symbolic-ref" "HEAD") "HEAD")))
